/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
 * Copyright (c) Sun Microsystems, Inc. 1998 All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Sun Microsystems, Inc.
 *
 * 4. The name of the Sun Microsystems, Inc nor may not be used to endorse or
 *      promote products derived from this software without specific prior
 *      written permission.
 *
 * SUN MICROSYSTEMS MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF THIS
 * SOFTWARE FOR ANY PARTICULAR PURPOSE.  The software is provided "as is"
 * without express or implied warranty of any kind.
 *
 * These notices must be retained in any copies of any part of this software.
 */

// #ident "@(#)mip.cc  1.4     98/08/21 SMI"

#include <template.h>
#include <mip.h>
#include <random.h>
#include <address.h>
#include <mobilenode.h>

#define IP_HEADER_SIZE	20

int hdr_ipinip::offset_;
static class IPinIPHeaderClass : public PacketHeaderClass {
public:
        IPinIPHeaderClass() : PacketHeaderClass("PacketHeader/IPinIP",
					    sizeof(hdr_ipinip*)) {
		bind_offset(&hdr_ipinip::offset_);
	}
} class_ipiniphdr;

static class MIPEncapsulatorClass : public TclClass {
public:
	MIPEncapsulatorClass() : TclClass("MIPEncapsulator") {}
	TclObject* create(int, const char*const*) {
		return (new MIPEncapsulator());
	}
} class_mipencapsulator;

MIPEncapsulator::MIPEncapsulator() : Connector(), mask_(0xffffffff),
	shift_(8), defttl_(32)
{
	bind("addr_", (int*)&(here_.addr_));
	bind("port_", (int*)&(here_.port_));
	bind("shift_", &shift_);
	bind("mask_", &mask_);
	bind("ttl_", &defttl_);
//	printf("MIPencapsulator::MIPencapsulator()\n"); //added by Leo, Mar-02-2002
}

void MIPEncapsulator::recv(Packet* p, Handler *h)
{

	Tcl& tcl = Tcl::instance();

//	printf ("MIPencapsulator::recv\n"); //added by Leo, Mar-08-2002
	hdr_ip* hdr = hdr_ip::access(p);
	hdr_ipinip **ppinhdr = (hdr_ipinip**)hdr_ipinip::access(p);
	if (--hdr->ttl_ <= 0) {
		/*
		 * XXX this should be "dropped" somehow.  Right now,
		 * these events aren't traced.
		 */
		hdr_ipinip *ptr = *ppinhdr, *temp;
		while (ptr != NULL) {
			temp = ptr;
			ptr = ptr->next_;
			delete temp;
		}
		*ppinhdr = NULL;
		Packet::free(p);
		return;
	}
	hdr_ipinip *inhdr = new hdr_ipinip;
	//int dst = ((hdr->dst() >> shift_) & mask_);
	int dst = Address::instance().get_nodeaddr(hdr->daddr());
	tcl.evalf("%s tunnel-exit %d", name_, dst);
	int te = atoi(tcl.result()); //the COA of the Mobile Host with the home address dst

	inhdr->next_ = *ppinhdr;
	*ppinhdr = inhdr;
	inhdr->hdr_ = *hdr;

//Added by Leo, Mar-10-2002
//This is to send Binding Update Message to the Correspendent Host
//If this part is commented, then it is the normal MIP NS without Binding Update message support
//When the HA's MIPEncapsulator received the packet, it should send the Bining Update message to the 
//Correspondent Host

	char* dstDomain = NULL;
	dstDomain = Address::instance().print_nodeaddr(dst); //Get the hierarchy address of the destination

//	printf("dstDomain:%s", dstDomain);
	
	tcl.evalf("%s nodeptr", name_);						//Invoke a TCL routine to return the node's hierarcy address

	char nodeDomain[10];
	strncpy(nodeDomain, tcl.result(), 10);				//Get the hierarchy address of the current node
	nodeDomain[9] = '\0';

//	printf("nodeDomain:%s", nodeDomain);
	
	if ((dstDomain[0]) == (nodeDomain[0]))				//If the destination and the current node in the same domain
	{													//Means the MIPEncapsulator is the encapsulator of the Home Agent
//		printf("In MIPEncapsulator!\n");
		tcl.evalf("%s send-bu %d %d %d", name_, hdr->saddr(), dst, te);			//Send out the Binding Update message, by invoking the 
																				//function of MIPBSAgent, through the TCL function
																				//Can be enhanced by directly call the method of MIPBSAgent
		int tmpRe = atoi(tcl.result());									
		if (tmpRe == 1)															//If the Binding Update message is sucessfully sent out
		{
			tcl.evalf("%s setCOA %d %d", name_, dst, te);						//Set the current Care-of Address and
			tcl.evalf("%s setCH %d %d", name_, dst, hdr->saddr());				//the Correspondent Host's address, which is equal to node id
		}
	}

	if (dstDomain != NULL)								//Recall the memory
		delete [] dstDomain;

//Ended Modification

	hdr->saddr() = here_.addr_;
	hdr->sport() = here_.port_;
	//hdr->dst() = addr_ & ~(~(nsaddr_t)0 << shift_) | (te & mask_) << shift_;;
	hdr->daddr() = te;
	hdr->dport() = 1;
	hdr->ttl() = defttl_;
	hdr_cmn::access(p)->size() += IP_HEADER_SIZE;

	target_->recv(p,h);

}

static class MIPDecapsulatorClass : public TclClass {
public:
	MIPDecapsulatorClass() : TclClass("Classifier/Addr/MIPDecapsulator") {}
	TclObject* create(int, const char*const*) {
		return (new MIPDecapsulator());
	}
} class_mipdecapsulator;

MIPDecapsulator::MIPDecapsulator() : AddressClassifier()
{
//	printf("MIPDecapsulator::MIPDecapsulator()\n"); //Added by Leo, Mar-03-2002
}

void MIPDecapsulator::recv(Packet* p, Handler *h)
{
	hdr_ipinip **ppinhdr = (hdr_ipinip **)hdr_ipinip::access(p);

	// restore inner header
	hdr_ip *pouthdr = hdr_ip::access(p);

//	char* dst = Address::instance().print_nodeaddr(pouthdr->daddr());	//Added by Leo, Mar-31-2002
//	char* src = Address::instance().print_nodeaddr(pouthdr->saddr());	//Added by Leo, Mar-31-2002

	assert(ppinhdr);
	hdr_ip *pinhdr = &(*ppinhdr)->hdr_;
	*ppinhdr = (*ppinhdr)->next_;
	*pouthdr = *pinhdr;

//	char* dst2 = Address::instance().print_nodeaddr((pinhdr)->daddr());	//Added by Leo, Mar-31-2002
//	char* src2 = Address::instance().print_nodeaddr((pinhdr)->saddr());	//Added by Leo, Mar-31-2002

	NsObject* link = find(p);
	// for mobilenodes use default-target which is probably the
	// RA. cannot use node_entry point instead, as hier address
	// of MH point to HA. hence hand decapsulated pkt directly
	// to RA.
	//tcl.evalf("warn %s", "sss");

	//printf("MIPDecapsulator::recv\n"); //Added by Leo, Mar-08-2002
	if (link == NULL || pinhdr->ttl_ <= 0) {
		/*
		 * XXX this should be "dropped" somehow.  Right now,
		 * these events aren't traced.
		 */
		hdr_ipinip *ptr = *ppinhdr, *temp;
		while (ptr != NULL) {
			temp = ptr;
			ptr = ptr->next_;
			delete temp;
		}
		*ppinhdr = NULL;
		delete pinhdr;
		Packet::free(p);

//Added by Leo, Mar-31-2002
/*	
		if (dst != NULL)
		{
			delete [] dst;
		}

		if (dst2 != NULL)
		{
			delete [] dst2;
		}

		if (src != NULL)
		{
			delete [] src;
		}

		if (src2 != NULL)
		{
			delete [] src2;
		}
*/
		return;
	}
	delete pinhdr;

//Added by Leo, Mar-31-2002
/*	
	hdr_cmn *ch = hdr_cmn::access(p);

	if (ch->timestamp() > 73)
	{
	
	printf("MIPDecapsulator packet outer IP: SA:%s, DA:%s\n", src, dst);
	printf("MIPDecapsulator packet inner IP: SA:%s, DA:%s\n", src2, dst2);

	}
//Ended Addition by Leo, Mar-31-2002

	if (dst != NULL)
	{
		delete [] dst;
	}

	if (dst2 != NULL)
	{
		delete [] dst2;
	}

	if (src != NULL)
	{
		delete [] src;
	}

	if (src2 != NULL)
	{
		delete [] src2;
	}
*/
	hdr_cmn::access(p)->size() -= IP_HEADER_SIZE;

	link->recv(p,h);

}
